# -*- coding: utf-8 -*-
"""
Created on Sun Feb 26 16:41:01 2017

@author: andi
"""

import visr
import panning
import pml
import objectmodel as om

from helper.baseTrigFunctions import cart2sph
from helper.real_sph_harmonics import allSphHarmRealACN

import numpy as np

class AllRadGainCalculator( visr.AtomicComponent ):
    """
       Panning gain computation component based on the AllRad algorithm.
    """
    def __init__(self, context, name, parent, *, # Asterisk has no effect but to force all subsequent
             numberOfObjects,
             regularArrayConfig,
             realArrayConfig,
             hoaOrder,
             orderWeights = None ):
        super( AllRadGainCalculator, self ).__init__(context, name, parent )

        self.hoaOrder = hoaOrder
        self.numberOfObjects = numberOfObjects
        self.numberOfOutputs = realArrayConfig.numberOfRegularLoudspeakers
        regularArraySize = regularArrayConfig.numberOfRegularLoudspeakers
        numAmbisonicSignals = (hoaOrder+1)**2

        self.objectInput = visr.ParameterInput( "objectVectorInput", self,
                                               parameterType=pml.ObjectVector.staticType,
                                               protocolType=pml.DoubleBufferingProtocol.staticType,
                                               parameterConfig=pml.EmptyParameterConfig() )
        self.objectInputProtocol = self.objectInput.protocolInput()

        matrixConfig = pml.MatrixParameterConfig( self.numberOfOutputs, self.numberOfObjects )
        self.coefficientOutput = visr.ParameterOutput( "gainOutput", self,
                                                pml.MatrixParameterFloat.staticType,
                                                pml.SharedDataProtocol.staticType,
                                                matrixConfig )
        self.coefficientOutputProtocol = self.coefficientOutput.protocolOutput()

        # Computation of the decoder matrix
        vbap = panning.VBAP( realArrayConfig )
        Lreg = regularArrayConfig.positions()
        Preg = np.zeros( ( self.numberOfOutputs, regularArraySize ),
                        dtype = np.float32 )
        for idx in range( regularArraySize ):
            tmpGains = np.zeros( ( self.numberOfOutputs), dtype=np.float32 )
            vbap.calculateGains( Lreg[idx,0], Lreg[idx,1], Lreg[idx,2], tmpGains )
            Preg[:,idx] = tmpGains

        az, el, r = cart2sph( Lreg )
        Yreg = allSphHarmRealACN( hoaOrder, 0.5*np.pi-el, az, dtype = Lreg.dtype )

        orderWeightsVec = np.ones( (numAmbisonicSignals), dtype = Lreg.dtype )
        if orderWeights is not None:
            rawWeights =  np.asarray(orderWeights, dtype = Lreg.dtype  )
            if rawWeights.size != hoaOrder + 1:
                raise ValueError( "The optional argument 'orderWeights' must contain hoaOrder+1 values." )
            for (order, weight ) in enumerate( rawWeights ):
                orderWeightsVec[order**2:(order+1)**2] = weight

        weightMtx = np.diag( orderWeightsVec )

        #self.decodeMtx = efl.BasicMatrixFloat(Preg @ (Yreg.T @ weightMtx) )
        scaleCoeff = (4.0*np.pi)/float(regularArraySize)
        self.decodeMtx = scaleCoeff * (Preg @ (Yreg.T @ weightMtx))

    def process( self ):
        if self.objectInputProtocol.changed():
            objectVector = self.objectInputProtocol.data()
            gainMtx = np.asarray(self.coefficientOutputProtocol.data())
            for obj in objectVector:
                if obj.type != om.ObjectType.PointSource:
                    continue
                outChannel = obj.channels[0]

                srcAz, srcEl, srcRadius =  cart2sph( np.asarray(obj.position) )
                b =allSphHarmRealACN( self.hoaOrder, 0.5*np.pi-srcEl, srcAz, dtype=np.float32 )

                gains = self.decodeMtx @ b
                gainMtx[:,outChannel] = gains

            self.objectInputProtocol.resetChanged()
