#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 17 14:22:59 2017

@author: af5u13
"""

import numpy as np
from scipy.spatial import ConvexHull

from collections import namedtuple
from copy import deepcopy

from helper.vectorFunctions import normalise, rotateToXYPlane, edgeNormalVectors

# VISR modules
import visr
import pml
import objectmodel

class VbapL1L2Panner( visr.AtomicComponent ):
    def __init__( self, context, name, parent,
                 numObjects, lspArray ):
        super().__init__( context, name, parent )
        self.L = normalise( lspArray.positions()[:lspArray.numberOfRegularLoudspeakers,:].T, norm=2, axis=0 )
        self.p = np.zeros( (numObjects, 3 ), dtype=np.float32 )
        self.objectIn = visr.ParameterInput( "objectVectorInput", self,
            pml.ObjectVector.staticType,
            pml.DoubleBufferingProtocol.staticType,
            pml.EmptyParameterConfig() )
        self.gainOut = visr.ParameterOutput( "gainOutput", self,
            pml.MatrixParameterFloat.staticType,
            pml.SharedDataProtocol.staticType,
            pml.MatrixParameterConfig(self.L.shape[1], numObjects ) )
        self.panner = VBAPL1L2Opt( self.L, normalisationNorm=2 )

    def process( self ):
        if self.objectIn.protocol.changed():
            self.objectIn.protocol.resetChanged()
            objVec = self.objectIn.protocol.data()
            gains = np.asarray( self.gainOut.protocol.data() )
            for obj in [o for o in objVec
             if isinstance( o, objectmodel.PointSource )]:
                self.p[obj.objectId,:] = obj.position
            gains[...] = self.panner.gains( self.p ).T

# from sphereGeometryPrimitives import polygonFromConstraints

# Graphical debugging
#from plotTesselation import plotTesselation, plotPoint, plotArc, plotGreatCircle
#import matplotlib.pyplot as plt
#from mpl_toolkits.mplot3d import Axes3D

# Create a named tuple data type.
Facet = namedtuple( 'Facet', [ 'edgeIndices', 'edgeNormals', 'lspIndices', 'inverse', 'centre' ] )

# TODO: Local definition possible?
IntersectionRecord = namedtuple( 'IntersectionRecord', ['vertexIndex', 'position'])

def triangleCentre( L ):
    rhs = np.ones( L.shape[0:-2]+(L.shape[-1],), dtype=L.dtype )
    cRaw = np.linalg.solve( L, rhs )
    cNorm = np.linalg.norm( cRaw, ord = 2, axis = -1 )
    cNormed = cRaw / cNorm[...,np.newaxis]
    return cNormed

class VBAPL1L2Opt:
    def __init__( self, lspArray, normalisationNorm=2 ): # , debugPlot = False ):

#        if debugPlot:
#            figure = plt.figure()
#            debugPlotAxis = figure.gca(projection='3d')
#        else:
#            debugPlotAxis = None

        if (lspArray.ndim != 2) or (not lspArray.shape[0] in [2,3] ):
            raise ValueError( 'The parameter "lspArray" must be a 2-D array with 2 or 3 as second dimension.' )
        self.normalisationNorm = normalisationNorm
        self.dim, self.numLoudspeakers = lspArray.shape

        L = normalise( lspArray.T, axis = -1 )

        # Members used during construction
        self.allVertices = [lVec for lVec in L ] # Store the vertices in a list
        self.additionalIndexStart = self.numLoudspeakers

        hull = ConvexHull( L )

        # %% Start with an initial tesselation of cyclic polygons.
        # TODO: Move to library function.
        tripletVertices = hull.points[hull.simplices]
        centres = triangleCentre( tripletVertices )

        distMtx = centres @ centres.T

        # Tolerance of dot products when searching for coplanarity.
        distEps = 1e-5 # TODO: Relate tolerance to angle

        allIndices = np.array( range(0,centres.shape[0]))

        # Data structure to collect the final 'facets' or polygons
        self.facets = []

        for idx in range( 0, centres.shape[0] ):
            # Check for special marker value whether this triplet has already been processed.
            if np.isnan( distMtx[idx,0] ):
                continue

            newFacet = set( hull.simplices[idx] )

            matchMask = distMtx[idx,idx+1:] >= (1.0-distEps )

            matchIndices = allIndices[idx+1:][matchMask]

            # Mark the rows of the processed other triangles as processed.
            distMtx[matchIndices,:] = np.NAN

            for extraTriangleIdx in matchIndices:
                extraTriangle = set( hull.simplices[extraTriangleIdx,...] )
                newFacet = newFacet.union( extraTriangle )

            # %% Compute a nonintersecting ordering (counterclockwise
            # on the face of the plane)
            normalVec = centres[idx,:]

            facetIndices = np.asarray(list(newFacet)) # Create a ndarray if ondices from a set
            vertices = L[facetIndices,:]

            # %% Rotate the normal plane to the x-y plane
            # First, move all the vectors to the plane perpendicular to the centre vector
            projFac = np.dot( vertices, normalVec )
            verticesInPlane = vertices - projFac[:,np.newaxis] * normalVec[np.newaxis,:]
            planeVertices = rotateToXYPlane( verticesInPlane, normalVec )

            planeAngles = np.arctan2( planeVertices[...,1], planeVertices[...,0])
            sortIdcs = np.argsort( planeAngles )
            # planeVerticesSorted = planeVertices[ sortIdcs, : ]
            indicesSorted = facetIndices[sortIdcs]

            # points = L[indicesSorted,: ]

            centre = np.asarray(centres[idx])
            # Pick one of the vertices for computing the radius (results are equal for all vertices)
            cosRadius = np.dot( centre, self.allVertices[indicesSorted[0]] )

            self.appendPolygon( list(indicesSorted), list(indicesSorted), centre, cosRadius ) # , debugPlotAxis )

        self.vertices = np.asarray( self.allVertices ).T
        del  self.allVertices


    def appendPolygon( self, vertexIndices, activeIndices, centre, cosRadius, posTolerance = 1e-6  ): # , debugPlotAxis = None
        """ Recursive method to partition a cyclic polygon into a set of elementary
            polygons with a fixed set of nonzero gains. """
        numVertices = len(vertexIndices) # Is this better: len(activeIndices)?
        vertexList = [self.allVertices[i] for i in vertexIndices]
        vertexMtx = np.asarray( vertexList )
        activeLsp = np.asarray( [self.allVertices[i] for i in activeIndices ] )

        # %% Shortcut solution for simple triplets:
        # This is not strictly necessary because the splitting algorithm will arrive at the same solution
        if numVertices <= 3:
            invMtx = np.linalg.pinv( activeLsp )
            # Facet = namedtuple( 'Facet', [ 'edgeIndices', 'edgeNormals', 'lspIndices', 'vertexIndices', 'inverse', 'centre' ] )
            edgeNormals = edgeNormalVectors( vertexMtx )
            self.facets.append( Facet( edgeIndices=list(vertexIndices),
                                       edgeNormals=edgeNormals,
                                       lspIndices=list(activeIndices),
                                       inverse=invMtx,
                                       centre=centre ) )
            return

        # Compute the 1st-order mode matrix
        # Note: This matrix is essentially singular.
        # Using double precision helps somehow (for the time being)
        # TODO: Change algorithm to avoid this potential numerical instability
        Y = np.concatenate( (activeLsp, np.ones( (len(activeIndices),1), dtype=np.float64 ) ), axis=1 ).T
        # For the time being, use the full list of loudspeakers for the constraints.
        # Y = np.concatenate( (allLsp, np.ones( (1, allLsp.shape[-1]), dtype=allLsp.dtype ) ), axis=0 )

        p0 = 1.0/cosRadius # Target pressure value (or value of the W mode),
        # determined by construction of VBAP as the maximum possible value for the correct particle velocity direction
        # under a nonnegativity constraint)

        # Transform matrix to get from the direction vector to the first-order mode vector with the (achivable,
        # optimal) pressure value p0
        T = np.concatenate( (np.identity(3), p0*np.reshape( centre, (1,3) ) ) )

        # Deprecated: direct pseudoinversion of a rank-deficient matrix
        Ypi = np.linalg.pinv( Y ) # Pseudoinverse matrix.
        Cold = normalise(Ypi @ T, axis = -1 ) # Constraint matrix of the polygon, containing unit vectors.

        # %% Attempt to compute the constraint matrix in a stable way.
        Tpi = np.linalg.pinv( T ) # Form the pseudoinverse of the transformation matrix
        # to apply the transformation of Y before the inversion.
        Ypr = Tpi @ Y
        CnewRaw = np.linalg.pinv( Ypr )
        Cnew = normalise( CnewRaw, axis=-1 )

        # Select the new solution (more stable).
        # TODO: Remove old computation in time.
        C = Cnew

        edgeNormals = edgeNormalVectors( vertexMtx )
        edgeLengths = np.arccos( np.sum( vertexMtx * vertexMtx[list(range(1,numVertices))+[0],:], axis = -1  ) ) # Effectively a row-wise dot product.

#        if debugPlotAxis is not None:
#            numEdges = vertexMtx.shape[0]
#            for idx in range( numEdges ):
#                plotGreatCircle( debugPlotAxis, edgeNormals[idx,:], color=[0.5,0.5,0.5], linewidth=0.5 )
#                plotArc( debugPlotAxis, vertexMtx[idx,:], vertexMtx[(idx+1)%numEdges,:], 'k-', linewidth=1.5 )
#            plt.pause(1)


        # numConstraints = vertices.shape[1]
        numConstraints = C.shape[0]
        splitFound = False # Flag signalling whether a intersecting condition has been found & therefore the loop has finished before completion.

        for constraintIdx in range( numConstraints ):
            constraintNormal = C[constraintIdx,:]

            # Dismiss any constraints that coincide with an existing edge.
            constraintDots = np.dot( edgeNormals, constraintNormal )
            # At the moment, we don't disambiguate between normal vectors that point into the same direction and diametral ones.
            # (If orientations are accounted for correcctly, the -1 case wpuld point to a feasible region just on the edge.)
            if np.any( np.abs( constraintDots >= 1.0 - posTolerance )):
                continue

#            if debugPlotAxis is not None:
#                plotGreatCircle( debugPlotAxis, constraintNormal, 'r-', linewidth=1 )
#                plt.pause(1)

            intersectionRecords = []

            for edgeIdx in range( numVertices ):

                intersectionUnNorm = np.cross( edgeNormals[edgeIdx,:], constraintNormal )
                intersection = normalise( intersectionUnNorm )
                intersections = np.stack( (intersection, -intersection), axis = 0 )

                # %% Simpler approach: Use dot product to determine angular distance, and the cross product to find the direction.
                angularDistancesDot = np.dot( intersections, vertexMtx[edgeIdx,:] )
                distanceDirectionVec = np.cross( vertexMtx[edgeIdx,:], intersections )
                distanceDirections = np.sign( np.dot( distanceDirectionVec, edgeNormals[edgeIdx,:] ) )
                # np.minimum is to prevent values slightler largert than 1 due to rounding, which lead to NaNs
                angularDistances = np.arccos( np.maximum(np.minimum(angularDistancesDot, 1.0 ), -1.0) ) * distanceDirections
                # Todo: exclude cases where the constraint circle coincides with the edge.
                # This is identical to 'grazing' the polygon, and for convex polygons that means there is
                # no intersection that needs to be accounted.

                validDistances = (angularDistances >= - posTolerance) # and (angularDistances <= edgeLengths[edgeIdx])
                validIdcs = validDistances.nonzero()[0]
                if len( validIdcs ) == 0:
                    continue
                minValidIdx = np.argmin( angularDistances[validIdcs] )
                minDist = angularDistances[validIdcs][minValidIdx]
                if minDist >= edgeLengths[edgeIdx] + posTolerance:
                    continue
                intersectIdx = validIdcs[minValidIdx]
                # Temporary measure to exclude matching the same point twice (if the intersection lies on an existing vertex)
                if (len( intersectionRecords ) > 0) and (np.linalg.norm( intersections[intersectIdx,:] - intersectionRecords[-1].position, ord=2 ) <= posTolerance):
                    continue
                intersectionRecords.append( IntersectionRecord( edgeIdx, intersections[intersectIdx,:] ) )

            # Temporary workaround: Check whether the last and the first intersection point coincide and delete in this case.
            # Other consecutive occurences (without wraparound) are detected in the loop.
            # TODO: Provide cleaner solution.
            if (len( intersectionRecords ) >= 2 ) and (np.linalg.norm( intersectionRecords[0].position - intersectionRecords[-1].position, ord=2 ) <= posTolerance):
                del intersectionRecords[-1]

            if len( intersectionRecords ) < 2: # Not intersecting or grazing.
                continue

            def removeDuplicatesCircular( vertexList ):
                """
                Replace sequences of identical elements by a single occurence.
                Similar to C++ std::unique, but takes care of wraparound between first and last element.
                """
                # TODO: 'Pythonise' later
                if len( vertexList ) < 2: # Duplicates make no sense here.
                    return vertexList
                newList = [vertexList[-1]]
                for el in vertexList:
                    if el != newList[-1]:
                        newList.append( el )
                return newList[1:]

            # TODO: maybe move the detection and removal of multiple occurences of the same intersection.
            # Note: this potentially adds new vertices to the list before it is clear that there are no errors.
            intersectVertexIndicesRaw  = [self.insertVertex( irec.position, tolerance = posTolerance ) for irec in intersectionRecords]
            intersectVertexIndices = removeDuplicatesCircular( intersectVertexIndicesRaw )
            if( len(intersectVertexIndices) > 2 ):
                raise ValueError( 'A convex polygon cannot be intersected more than twice by a plane.' )
            else:

                vertexIndices1 = removeDuplicatesCircular( vertexIndices[0:intersectionRecords[0].vertexIndex+1] + intersectVertexIndices + vertexIndices[intersectionRecords[1].vertexIndex+1:] )
                vertexIndices2 = removeDuplicatesCircular( intersectVertexIndices[::-1] + vertexIndices[intersectionRecords[0].vertexIndex+1 : intersectionRecords[1].vertexIndex+1] )

                # %% Determine in which new polygon the current constraint is 'avtive'.
                # Apparently, this does not follow trivially from the intersection algorithm
                #
                vertices1Mtx = np.asarray( [self.allVertices[i] for i in vertexIndices1] )
                vertices2Mtx = np.asarray( [self.allVertices[i] for i in vertexIndices2] )

                vert1ActiveDot = np.dot( vertices1Mtx, constraintNormal )
                vert2ActiveDot = np.dot( vertices2Mtx, constraintNormal )

                vert1Active = np.all( vert1ActiveDot >= - posTolerance )
                vert2Active = np.all( vert2ActiveDot >= - posTolerance )

                if not (vert1Active ^ vert2Active ):
                    raise RuntimeError( 'The constraint must be active for exactly one of the created sub-polygons' )
#                activeIndices1 = activeIndices # Initialise bothe lists with the current constraints set.
#                activeIndices2 = activeIndices

                activeIndices1 = deepcopy( activeIndices )# Initialise bothe lists with the current constraints set.
                activeIndices2 = deepcopy( activeIndices )


                deletedConstraint = activeIndices[constraintIdx]
                if vert1Active:
                    # del activeIndices2[ constraintIdx ]
                    activeIndices2.remove( deletedConstraint )
                else:
                    # del activeIndices1[ constraintIdx ]
                    activeIndices1.remove( deletedConstraint )
#                if vert1Active:
#                    activeIndices1 = activeIndices
#                    activeIndices2 = activeIndices - {constraintIdx} # this is for the polygon where the current constraint is fulfilled
#                else:
#                    activeIndices1 = activeIndices - {constraintIdx} # this is for the polygon where the current constraint is fulfilled
#                    activeIndices2 = activeIndices

#                if  debugPlotAxis is not None:
#                     debugPlotAxis.clear()
                self.appendPolygon( vertexIndices1, activeIndices1, centre, cosRadius, posTolerance = posTolerance ) #, debugPlotAxis = debugPlotAxis )

#                if  debugPlotAxis is not None:
#                     debugPlotAxis.clear()
                self.appendPolygon( vertexIndices2, activeIndices2, centre, cosRadius, posTolerance = posTolerance ) #, debugPlotAxis = debugPlotAxis )

                splitFound = True
                break

        if not splitFound:
            invMtx = np.linalg.pinv( activeLsp )
            # Facet = namedtuple( 'Facet', [ 'edgeIndices', 'edgeNormals', 'lspIndices', 'vertexIndices', 'inverse', 'centre' ] )
            edgeNormals = edgeNormalVectors( vertexMtx )
            self.facets.append( Facet( edgeIndices=list(vertexIndices),
                                       edgeNormals=edgeNormals,
                                       lspIndices=list(activeIndices),
                                       inverse=invMtx,
                                       centre=centre ) )

    def insertVertex( self, vertex, tolerance = 1.0e-10 ):
        """
            Register a new vertex coordinate (if it not already exists).
            Return the index to the existing or newly inserted vertex position.
        """
        for idx, pos in enumerate(self.allVertices):
            if np.linalg.norm( pos - vertex, ord=2 ) <= tolerance:
                return idx
        newIndex = len( self.allVertices )
        self.allVertices.append( vertex )
        return newIndex

    def facets( self ):
        return self.facets

    def vertices( self ):
        return self.allVertices

    def lspVertices( self ):
        return self.allVertices[:,0:self.numLoudspeakers]

    def virtualVertices( self ):
        return self.allVertices[:,self.numLoudspeakers:]

    def gains( self, p ):
        sourceShape = p.shape[0:-1]
        pNormed = normalise( p, axis = -1 )

        g = np.zeros( sourceShape + (self.numLoudspeakers,), dtype=p.dtype )

        numFacets = len(self.facets)

        activeIndicator = np.zeros( sourceShape + (numFacets,), dtype=p.dtype )

        for facetIdx in range( 0, len(self.facets) ):
            constraintMtx = self.facets[facetIdx].edgeNormals
            gRaw = np.matmul( constraintMtx, pNormed.T )
            activeIndicator[...,facetIdx] = np.min( gRaw, axis=0)

        activeFacetIdx = np.argmax( activeIndicator, axis = -1 )

        # At the moment this only works for 1D arrays of sources (2D matrix)
        # TODO: Allow arbitrary dimensions for the source vector.
        for sourceIdx in range(pNormed.shape[0]):
            facet = self.facets[ activeFacetIdx[sourceIdx] ]
            gRaw = np.dot( pNormed[sourceIdx,:], facet.inverse )
            g[ sourceIdx, facet.lspIndices ] = gRaw

        if self.normalisationNorm == 0:
            return g
        else:
            gNorm = np.linalg.norm( g, ord=self.normalisationNorm, axis=-1 )
            gNormed = g / 1.0/gNorm[...,np.newaxis]
            return gNormed
